/*
 * Copyright (c) 1992-1994 Michael A. Cooper.
 * This software may be freely distributed provided it is not sold for 
 * profit and the author is credited appropriately.
 */

#ifndef lint
static char *RCSid = "$Id: obp.c,v 1.57 1996/02/14 20:51:03 mcooper Exp $";
#endif

/*
 * Open Boot PROM (OBP) routines
 */

#include "defs.h"

#if	defined(HAVE_OPENPROM)

#if SUNOS >= 5
#	include <sys/openprom.h>
#	include <sys/ddi.h>
#	include <sys/sunddi.h>
#	include <sys/ddi_impldefs.h>
#	define DEVI_CHILD(dev)		dev->devi_child
#	define DEVI_SIBLING(dev)	dev->devi_sibling
#	define DEVI_UNIT(dev)		dev->devi_instance
#	define DEVI_NODEID(dev)		dev->devi_nodeid
#	define DEVI_EXISTS(dev)		(dev->devi_addr && dev->devi_ops)
#else
#	include <sun/openprom.h>
#	define DEVI_CHILD(dev)		dev->devi_slaves
#	define DEVI_SIBLING(dev)	dev->devi_next
#	define DEVI_UNIT(dev)		dev->devi_unit
#	define DEVI_NODEID(dev)		0 /* No such member */
#	define DEVI_EXISTS(dev)		dev->devi_driver
#endif	/* SUNOS == 5 */

#define OBP_IS_BOOL(tp)		(tp->obpt_key[strlen(tp->obpt_key)-1] == '?' \
				 || tp->obpt_value[0] == CNULL)
/*
 * OBPio_t is used to do I/O with openprom.
 */
typedef union {
    char			opio_buff[OPROMMAXPARAM];
    struct openpromio		opio_oprom;
} OBPio_t;

/*
 * OBPtable_t is used to store properties for OBP node.
 */
typedef struct _obptable {
    char		        obpt_key[OPROMMAXPARAM];
    char		        obpt_value[OPROMMAXPARAM];
    struct _obptable	       *obpt_next;
} OBPtable_t;

/*
 * Various declarations
 */
extern char		 	OpenPROMSYM[];
typedef struct dev_info		OBPdev_info_t;
static int 			OBPKtraverse();
static kvm_t		       *kd = NULL;
static long			CpuClockFreq = 0;

/*
 * Get the root (top) node from the OBP.
 */
static OBPdev_info_t *OBPKgetRoot()
{
    struct nlist	       *nlptr;
    static OBPdev_info_t 	Root, *PtrRoot = NULL;
    u_long 			Addr;

    /*
     * See if we've already gotten Root
     */
    if (PtrRoot)
	return(&Root);

    if (!kd)
	if (!(kd = KVMopen()))
	    return((OBPdev_info_t *) NULL);

    if ((nlptr = KVMnlist(kd, OpenPROMSYM, (struct nlist *)NULL, 0)) == NULL)
	return((OBPdev_info_t *) NULL);

    if (CheckNlist(nlptr))
	return((OBPdev_info_t *) NULL);

    /*
     * Read pointer to "top_devinfo" from kernel
     */
    Addr = nlptr->n_value;
    if (KVMget(kd, Addr, (char *) &PtrRoot, 
	       sizeof(struct dev_info *), KDT_DATA)) {
	if (Debug) Error("Cannot read openprom devinfo pointer from kernel");
	return((OBPdev_info_t *) NULL);
    }

    if (KVMget(kd, (u_long)PtrRoot, (char *)&Root, 
		sizeof(OBPdev_info_t), KDT_DATA)) {
	if (Debug) Error("Cannot read openprom devinfo root from kernel");
	return((OBPdev_info_t *) NULL);
    }

    return(&Root);
}

/*
 * Build device tree by reading the Open Boot PROM from the kernel.
 */
static int OBPKbuild(TreePtr, SearchNames)
    DevInfo_t 		      **TreePtr;
    char		      **SearchNames;
{
    OBPdev_info_t	       *Root;

    if (!(Root = OBPKgetRoot()))
	return(-1);

    return(OBPKtraverse(Root, NULL, TreePtr, SearchNames));
}

/*
 * Cleanup an OBP node name.
 */
extern char *OBPcleanName(Name, FullClean)
    char		       *Name;
    int				FullClean;
{
    register char	       *cp;
    register char	       *bp;
    register char	       *end;
    register int		len;
    char		       *Buffer;
    char		       *shortman;

    if (!Name || !Name[0])
	return((char *) NULL);

    /* XXX - Assume the new string won't be longer than the original. */
    Buffer = (char *) xmalloc(strlen(Name)+1);

    if (FullClean) {
	/*
	 * Skip over initial wording up to ','.  e.g. "SUNW,4/25" -> "4/25".
	 */
	if (cp = strchr(Name, ','))
	    Name = ++cp;

	for (cp = Name, bp = Buffer; cp && *cp; ++cp) {
	    /*
	     * Change '-' to ' '
	     * Change '_' to '/' (for SunOS 5.x)
	     * Skip ','
	     */
	    if (*cp == '-')
		*bp++ = ' ';
	    else if (*cp == '_')
		*bp++ = '/';
	    else if (*cp != ',')
		*bp++ = *cp;

	    *bp = CNULL;
	}

    } else {
	Buffer = strdup(Name);
	/* Make bp point at end of Buffer */
	bp = Buffer;
	bp += strlen(Buffer);
    }

    /*
     * Remove all trailing spaces.
     * bp points to end of Buffer at this point.
     */
    end = bp;
    for (cp = end; cp > Buffer && *cp == ' '; --cp);
    if (cp != end && cp && *++cp == ' ')
	*cp = CNULL;

    /*
     * Remove manufacturer part of name.
     * If shortman is "Sun" and Name is "Sun 4/25",
     * change Name to "4/25".
     */
    cp = Buffer;
    if (shortman = GetManShort()) {
	len = strlen(shortman);
	if (EQN(cp, shortman, len) && (int)strlen(cp) > len+2 && 
	    cp[len] == ' ') {
	    cp += strlen(shortman);
	    while (cp && *cp && *cp == ' ')
		++cp;
	}
    }

    return(cp);
}

/*
 * Check an OpenPROM device.
 */
static int OBPKcheckDevice(OBPDevInfo, Parent, TreePtr, SearchNames)
    OBPdev_info_t		*OBPDevInfo;
    OBPdev_info_t		*Parent;
    DevInfo_t 		       **TreePtr;
    char		       **SearchNames;
{
    static DevData_t 		 DevData;
    DevInfo_t 			*DevInfo;	

    /* Make sure devdata is clean */
    memset((void *)&DevData, 0, sizeof(DevData));
    DevData.DevUnit 		= -1;
    DevData.Slave 		= -1;
    DevData.CtlrUnit 		= -1;
    DevData.DevNum 		= -1;

    /* Set what we know */
    if (OBPDevInfo && DEVI_NODEID(OBPDevInfo) != -1)
	DevData.NodeID = DEVI_NODEID(OBPDevInfo);

    if (OBPDevInfo && OBPDevInfo->devi_name && OBPDevInfo->devi_name[0]) {
	DevData.DevName = OBPcleanName(OBPDevInfo->devi_name, FALSE);
	/*
	 * If we have a parent, we're not the ROOT node.
	 * Otherwise we are the ROOT node.
	 * If so, we set the unit number to -1
	 * to avoid a bug in SunOS 4.x.
	 */
	if (Parent)
	    DevData.DevUnit = DEVI_UNIT(OBPDevInfo);
	else
	    DEVI_UNIT(OBPDevInfo) = -1;
    }
    if (Parent && Parent->devi_name && Parent->devi_name[0]) {
	DevData.CtlrName = OBPcleanName(Parent->devi_name, FALSE);
	DevData.CtlrUnit = DEVI_UNIT(Parent);
    }
    /* 
     * OpenPROM nodes that have a driver ALWAYS exist.
     * Some nodes may exist, without a driver, however.
     */
    if (DEVI_EXISTS(OBPDevInfo))
	DevData.Flags |= DD_IS_ALIVE;

#if	SUNOS == 5
    /*
     * Get device number
     */
    if (OBPDevInfo->devi_minor) {
	static struct ddi_minor_data	MinorData;

	if (KVMget(kd, (u_long) OBPDevInfo->devi_minor, (char *) &MinorData,
		   sizeof(struct ddi_minor_data), KDT_DATA) == 0 &&
	    MinorData.type == DDM_MINOR)
	    DevData.DevNum = MinorData.ddm_dev;
    }
#endif	/* SUNOS == 5 */

    if (Debug) {
	printf("OPENPROM: Found \"%s\" Node %d (Unit %d Dev %d) ",
	       ARG(DevData.DevName), OBPDevInfo->devi_nodeid,
	       DevData.DevUnit, DevData.DevNum);
	printf("on \"%s\" (Unit %d)%s\n",
	       ARG(DevData.CtlrName), DevData.CtlrUnit,
	       (DevData.Flags & DD_IS_ALIVE) ? " [ALIVE]" : "");
    }

#if	SUNOS == 4
    /*
     * XXX Avoid hanging SLC & ELC machines which don't have floppy disks.
     */
    if (DevData.DevName && EQ(DevData.DevName, "fd")) {
#if	defined(CPU_SUN4C_20)
	if (CpuType == CPU_SUN4C_20) return;
#endif	/* CPU_SUN4C_20 */
#if	defined(CPU_SUN4C_25)
	if (CpuType == CPU_SUN4C_25) return;
#endif	/* CPU_SUN4C_25 */
#if	defined(CPU_SUN4C_30)
	if (CpuType == CPU_SUN4C_30) return;
#endif	/* CPU_SUN4C_30 */
    }
#endif	/* SUNOS == 4 */

    /* Probe and add device */
    if (TreePtr && (DevInfo = (DevInfo_t *) ProbeDevice(&DevData, 
							TreePtr, SearchNames)))
	AddDevice(DevInfo, TreePtr, SearchNames);
}

/*
 * Recursively traverse and descend the OpenPROM devinfo tree.
 */
static int OBPKtraverse(OBPDevInfo, Parent, TreePtr, SearchNames)
    OBPdev_info_t		*OBPDevInfo;
    OBPdev_info_t		*Parent;
    DevInfo_t 		       **TreePtr;
    char		       **SearchNames;
{
    static char 		 Name[BUFSIZ];
    OBPdev_info_t		*Ptr;

    /*
     * If node name is a valid pointer, read the name from kernel space
     * and call probe routine to handle checking the device.
     */
    if (OBPDevInfo->devi_name) {
	if (KVMget(kd, (u_long) OBPDevInfo->devi_name, (char *) Name, 
		   sizeof(Name), KDT_STRING)) {
	    if (Debug) Error("Cannot read openprom device name.");
	    Name[0] = CNULL;
	} else {
	    OBPDevInfo->devi_name = (char *) strdup(Name);
	    OBPKcheckDevice(OBPDevInfo, Parent, TreePtr, SearchNames);
	}
    }

    /*
     * If this node has children, read the child data from kernel space
     * and DESCEND.
     */
    if (DEVI_CHILD(OBPDevInfo)) {
	Ptr = (OBPdev_info_t *) xcalloc(1, sizeof(OBPdev_info_t));
	if (KVMget(kd, (u_long) DEVI_CHILD(OBPDevInfo), (char *) Ptr,
		   sizeof(OBPdev_info_t), KDT_DATA)) {
	    Error("Cannot read openprom slave data for %s.", Name);
	} else {
	    DEVI_CHILD(OBPDevInfo) = (OBPdev_info_t *) Ptr;
	    OBPKtraverse(DEVI_CHILD(OBPDevInfo), OBPDevInfo, 
			 TreePtr, SearchNames);
	}
    }

    /*
     * If this node has a sibling, read the sibling data from kernel space
     * and TRAVERSE.
     */
    if (DEVI_SIBLING(OBPDevInfo)) {
	Ptr = (OBPdev_info_t *) xcalloc(1, sizeof(OBPdev_info_t));
	if (KVMget(kd, (u_long) DEVI_SIBLING(OBPDevInfo), (char *) Ptr,
		   sizeof(OBPdev_info_t), KDT_DATA)) {
	    Error("Cannot read openprom next data for %s.", Name);
	} else {
	    DEVI_SIBLING(OBPDevInfo) = (OBPdev_info_t *) Ptr;
	    OBPKtraverse(DEVI_SIBLING(OBPDevInfo), Parent, 
			 TreePtr, SearchNames);
	}
    }

    return(0);
}

/*
 * Get next OBP node id using "what" as what to get.
 */
static OBPnodeid_t OBPIOnext(NodeID, obp_fd, What)
    OBPnodeid_t		        NodeID;
    int				obp_fd;
    int				What;
{
    OBPio_t			opio;
    struct openpromio	       *op;
    int			       *idptr;

    op = &(opio.opio_oprom);
    idptr = (OBPnodeid_t *) (op->oprom_array);
    memset((void *)opio.opio_buff, 0, sizeof(opio.opio_buff));
    *idptr = NodeID;
    op->oprom_size = sizeof(opio.opio_buff);

    if (ioctl(obp_fd, What, op) < 0) {
	if (Debug) Error("OBP ioctl %s failed: %s.", 
			 (What == OPROMNEXT) ? "OPROMNEXT" : "OPROMCHILD", 
			 SYSERR);
	return(0);
    }

    return(*((OBPnodeid_t *)op->oprom_array));
}

/*
 * Determine if opvalue is a string or not.
 */
static int IsString(opvalue, opsize)
    char		       *opvalue;
    int				opsize;
{
    register int		i;

    for (i = 0; i < opsize - 1; ++i) {
	if (opvalue[i] == CNULL)
	    return(0);
	if (iscntrl(opvalue[i]) || !isascii(opvalue[i]))
	    return(0);
    }

    return(1);
}

static char *GetLongValStr(String)
    char		       *String;
{
    static char			buff[BUFSIZ];
    long			lval;
    long		       *lptr;
    long			nval;

    lval = strtol(String, (char **)NULL, 0);
    lptr = (long *) &String[0];
    nval = *lptr;
    (void) sprintf(buff, "%u", (u_long) nval);

    return(buff);
}

/*
 * Decode an OBP value into a printable string.
 */
static char *OBPdecodeval(Value, Size)
    char		       *Value;
    int				Size;
{
    static char			Buff[2*BUFSIZ];
    static char			Buff2[2];
    register int		i;
    int				OpMask = 0xff;

    if (Size == 0 || !Value)
	return((char *) NULL);

    if (IsString(Value, Size)) {
	(void) strcpy(Buff, Value);
	return(Buff);
    } else {
	if (Size > 4) {
	    if (Size > OpMask)
		return((char *) NULL);
	    (void) strcpy(Buff, "0x");
	    for (i = 0; i < Size; ++i) {
		if (i > 0 && ((i % 4) == 0))
		    (void) strcat(Buff, ".0x");
		(void) sprintf(Buff2, "%02x", Value[i] & OpMask);
		(void) strcat(Buff, Buff2);
	    }
	} else
	    (void) strcpy(Buff, GetLongValStr(Value));
	return(Buff);
    }
}

/*
 * Get the property value of propname from the OBP.
 */
static char *OBPgetpropval(propname, obp_fd)
    char		       *propname;
    int				obp_fd;
{
    static OBPio_t		opio;
    struct openpromio	       *op;

    op = &(opio.opio_oprom);
    op->oprom_size = sizeof(opio.opio_buff);
    (void) strcpy(op->oprom_array, propname);

    if (ioctl(obp_fd, OPROMGETPROP, op) < 0) {
	if (Debug) Error("OBP ioctl OPROMGETPROP failed for '%s': %s.", 
			 propname, SYSERR);
	return((char *)NULL);
    }

    if (op->oprom_size == -1) {
	if (Debug) Error("OBP no data available for '%s'.", propname);
	return((char *)NULL);
    }

    /*
     * op->oprom_array points at opio.opio_buff where the results
     * really live.
     */
    return(OBPdecodeval(op->oprom_array, op->oprom_size));
}

/*
 * Create and copy a OBPtable_t
 */
static OBPtable_t *CopyOBPtable(old)
    OBPtable_t	       *old;
{
    OBPtable_t	       *new;

    new = (OBPtable_t *) xmalloc(sizeof(OBPtable_t));
    memcpy((void *)new, (void *)old, sizeof(OBPtable_t));

    return(new);
}

/*
 * Get all OBP properteries for the current OBP node and return
 * a table of the results.
 */
static OBPtable_t *OBPIOgetproptab(obp_fd)
    int				obp_fd;
{
    static OBPio_t		opio;
    struct openpromio	       *op;
    static OBPtable_t	       *OBPtable = NULL;
    static OBPtable_t		tab;
    register OBPtable_t        *tptr, *tabptr;
    register char	       *cp;
    register int		found;

    op = &(opio.opio_oprom);
    memset((void *)opio.opio_buff, 0, sizeof(opio.opio_buff));

    /*
     * Mark all previously allocated table entries as available
     */
    if (OBPtable)
	for (tptr = OBPtable; tptr; tptr = tptr->obpt_next)
	    tptr->obpt_key[0] = tptr->obpt_value[0] = CNULL;

    for ( ; ; ) {
	/*
	 * Get next property
	 */
	op->oprom_size = sizeof(opio.opio_buff);
	if (ioctl(obp_fd, OPROMNXTPROP, op) < 0) {
	    if (Debug) Error("OBP ioctl OPROMNXTPROP failed: %s.", SYSERR);
	    return((OBPtable_t *) NULL);
	}

	/*
	 * We're done
	 */
	if (!op->oprom_size)
	    break;

	if (op->oprom_size >= sizeof(tab.obpt_key)) {
	    if (Debug) Error("Openprom key \"%s\" too large for buffer.",
			     op->oprom_array);
	    continue;
	}
	(void) strcpy(tab.obpt_key, op->oprom_array);
	tab.obpt_value[0] = CNULL;

	/*
	 * Get the property value
	 */
	if (cp = OBPgetpropval(op->oprom_array, obp_fd)) {
	    if (strlen(cp) >= sizeof(tab.obpt_value)) {
		if (Debug) Error("Openprom value \"%s\" too large for buffer.",
				 cp);
		continue;
	    }
	    (void) strcpy(tab.obpt_value, cp);
	}

	/*
	 * Add this entry to the linked list.
	 * We try to use all previously allocated members of the linked
	 * list.  If none are available, we create a new member and add
	 * it to the list.  Allocated, but unused members have empty
	 * obpt_key and obpt_value strings.
	 */
	tab.obpt_next = NULL;
	if (OBPtable) {
	    for (tptr = OBPtable, found = FALSE; 
		 tptr && tptr->obpt_next && !found;
		 tptr = tptr->obpt_next) {

		/*
		 * See if this is an unused, but allocated entry
		 */
		if (tptr->obpt_key[0] == CNULL &&
		    tptr->obpt_value[0] == CNULL) {
		    tab.obpt_next = tptr->obpt_next;
		    memcpy((void *) tptr, (void *) &tab, sizeof(tab));
		    found = TRUE;
		}
	    }
	    /*
	     * If we didn't find an unused entry slot, then create a new one
	     */
	    if (!found)
		tptr->obpt_next = CopyOBPtable(&tab);
	} else
	    OBPtable = CopyOBPtable(&tab);
    }

    return(OBPtable);
}

/*
 * Get the CPU model.
 *
 * The model is based upon the type of CPU and the clock frequency
 * of the CPU.
 */
static char *OBPgetCPUmodel(cputype, OBPtable, clockfreqptr)
    char		       *cputype;
    OBPtable_t		       *OBPtable;
    long		       *clockfreqptr;
{
    register OBPtable_t	       *tptr;
    long 			clockfreq = 0;
    register int		i;
    static char			Buff[BUFSIZ];
    Define_t		       *Def;

    /*
     * See if this node has it's own clock frequency
     */
    for (tptr = OBPtable; tptr; tptr = tptr->obpt_next) {
	if (!tptr->obpt_key[0])
	    break;

	if (EQ(tptr->obpt_key, OBPK_CLOCKFREQ))
	    clockfreq = strtol(tptr->obpt_value, (char **)NULL, 0);
    }

    /*
     * No individual clock frequency, so we use the "global"
     * clock freqency that (hopefully) was set earlier.
     */
    if (!clockfreq)
	clockfreq = CpuClockFreq;

    /*
     * Set the frequency we're using so our caller can use it.
     */
    if (clockfreq && clockfreqptr)
	*clockfreqptr = clockfreq;

    /*
     * Scan the CPU types table looking for an entry that has
     * the same type name and the same clock frequency.
     */
    for (Def = DefGetList(DL_CPU); Def; Def = Def->Next) {
	if (EQ(Def->KeyStr, cputype)) {
	    if (Def->KeyNum == 0) {
		(void) sprintf(Buff, "%s %s", 
			       FreqStr(clockfreq), Def->ValStr1);
		return(strdup(Buff));
	    } else if ((Def->KeyNum == -1) ||
		       (clockfreq / MHERTZ) == Def->KeyNum)
		return(Def->ValStr1);
	}
    }

    return((char *)NULL);
}

/*
 * Expand and OBP key string into something more easily read.
 */
static char *OBPexpandKey(string)
    char		       *string;
{
    register char	       *cp;
    register int		len;
    static char			buff[BUFSIZ];

    if (!string)
	return((char *)NULL);

    (void) strcpy(buff, string);

    /* Capitolize first letter */
    buff[0] = toupper(buff[0]);

    for (cp = &buff[0]; cp && *cp; ++cp) {
	if (*cp == '-' || *cp == '_') {
	    *cp++ = ' ';
	    if (*cp)
		*cp = toupper(*cp);
	}
    }

    /* Remove ending '?' */
    len = strlen(buff);
    if (buff[len-1] == '?')
	buff[len-1] = CNULL;

    return(buff);
}

/*
 * Get a nice informative string describing OBPtab information and
 * set the info into DevInfo.
 */
static void OBPsetOBPinfo(OBPtab, DevInfo)
    OBPtable_t		       *OBPtab;
    DevInfo_t		       *DevInfo;
{
    char		       *ExpName = NULL;
    char		       *Val = NULL;
    Define_t		       *Def;

    /*
     * Try to see if this is one of the types of OBP info
     * that we want to use a more descriptive text than
     * the key itself.
     */
    Def = DefGet(DL_OBP, OBPtab->obpt_key, -1, 0);
    if (Def && Def->ValStr1)
	ExpName = Def->ValStr1;

    if (!ExpName)
	ExpName = OBPexpandKey(OBPtab->obpt_key);

    if (!OBP_IS_BOOL(OBPtab)) {
	if (EQ(OBPtab->obpt_key, OBPK_CLOCKFREQ))
	    Val = strdup(FreqStr(strtol(OBPtab->obpt_value, 
					(char **)NULL, 0)));
	else
	    Val = OBPtab->obpt_value;
    }

    if (OBP_IS_BOOL(OBPtab))
	AddDevDesc(DevInfo, ExpName, "Has", DA_APPEND);
    else
	AddDevDesc(DevInfo, Val, ExpName, DA_APPEND);
}

/*
 * Fix an OBP node name to be compatable with the
 * conventions used in SunOS 5.x.
 *
 * i.e. All '/' characters are mapped to '_'.
 */
static char *OBPfixNodeName(string)
    char		       *string;
{
    register char	       *cp;
    static char			buff[BUFSIZ];

    (void) strcpy(buff, string);
    cp = buff;
    while (cp = strchr(cp, '/'))
	*cp++ = '_';

    return(buff);
}

/*
 * Get (make) name of a CPU
 */
extern char *GetCpuName(ncpus, cpuid)
    int			       *ncpus;
    int				cpuid;
{
    static char			buff[BUFSIZ];

    (void) sprintf(buff, "cpu%d", (cpuid >= 0) ? cpuid : (*ncpus)++);

    return(buff);
}

/*
 * Probe a CPU using OBP info.
 */
extern DevInfo_t *OBPprobeCPU(NodeID, ParentID, OBPtable, TreePtr, SearchNames)
    OBPnodeid_t			NodeID;
    OBPnodeid_t			ParentID;
    OBPtable_t		       *OBPtable;
    DevInfo_t		      **TreePtr;
    char		      **SearchNames;
{
    register OBPtable_t	       *tptr;
    DevInfo_t		       *DevInfo;
    DevInfo_t		       *fdev;
    register char	       *cp;
    register int		i;
    static char			buff[BUFSIZ];
    static int			ncpus = 0;
    int				cpuid = -1;
    char		       *cpumodel = NULL;
    long			clockfreq = 0;

    if (!(DevInfo = NewDevInfo(NULL)))
	return((DevInfo_t *)NULL);

    if (Debug) {
	(void) sprintf(buff, "Node ID is %d", NodeID);
	AddDevDesc(DevInfo, buff, NULL, DA_APPEND);
    }

    for (tptr = OBPtable; tptr; tptr = tptr->obpt_next) {
	if (!tptr->obpt_key[0])
	    break;

	/*
	 * See if this is the CPU type
	 */
	if (EQ(OBPK_NAME, tptr->obpt_key)) {
	    cpumodel = tptr->obpt_value;
	    DevInfo->Model = OBPgetCPUmodel(cpumodel, OBPtable, &clockfreq);
	    continue;
	} else if (EQ(tptr->obpt_key, OBPK_CPUID))
	    cpuid = (int) strtol(tptr->obpt_value, (char **)NULL, 0);
	else if (EQ(tptr->obpt_key, OBPK_DEVTYPE) ||
		 EQ(tptr->obpt_key, OBPK_CLOCKFREQ))
	    continue;

	OBPsetOBPinfo(tptr, DevInfo);
    }

    if (cp = GetCpuName(&ncpus, cpuid))
	DevInfo->Name = strdup(cp);

    if (SearchNames && !SearchCheck(DevInfo->Name, SearchNames))
	return((DevInfo_t *) NULL);

    if (cpumodel) {
	if (clockfreq)
	    (void) sprintf(buff, "%s %s", FreqStr(clockfreq), cpumodel);
	else
	    (void) strcpy(buff, cpumodel);
	AddDevDesc(DevInfo, buff, NULL, DA_INSERT|DA_PRIME);
	/*
	 * This must be an unknown type of CPU
	 */
	if (!DevInfo->Model)
	    DevInfo->Model = PrimeDesc(DevInfo);
    }

    DevInfo->Type = DT_CPU;
    DevInfo->NodeID = NodeID;

    /*
     * If no device by our name exists try to find a device
     * with a name which matches our CPU Model.  This should
     * be the root device node.  If so, make our device name
     * match the root device node name, so AddDevice() will 
     * add our info to the root device node later.
     */
    if (TreePtr && *TreePtr &&
	!FindDeviceByName(DevInfo->Name, *TreePtr)) {

	fdev = FindDeviceByName(cpumodel, *TreePtr);
	if (!fdev) {
	    if (Debug) printf("Cannot find CPU device <%s>\n", cpumodel);
	    cp = OBPfixNodeName(cpumodel);
	    if (!EQ(cp, cpumodel))
		fdev = FindDeviceByName(cp, *TreePtr);
	} else
	    if (Debug) printf("Found CPU device <%s>\n", cpumodel);
	if (fdev && !fdev->Master)
	    DevInfo->Name = fdev->Name;
    }

    return(DevInfo);
}

/*
 * Get part description information using a model (part) name.
 */
static Define_t *OBPGetPart(PartName)
    char		       *PartName;
{
    Define_t		       *PartList;
    register Define_t	       *Ptr;
    register int		i;
    register char	       *cp;

    PartList = DefGetList(DL_PART);
    if (!PartList) {
	if (Debug) Error("No Part List found.");
	return((Define_t *) NULL);
    }

    for (Ptr = PartList; Ptr; Ptr = Ptr->Next) {
	if (EQN(Ptr->KeyStr, PartName, strlen(Ptr->KeyStr)))
	    return(Ptr);
	/* Try it without the manufacturer part of part name */
	if ((cp = strchr(Ptr->KeyStr, ',')) && 
	    EQN(++cp, PartName, strlen(cp)))
	    return(Ptr);
    }

    return((Define_t *) NULL);
}

/*
 * This is for the SPARCcenter 2000 and SPARCserver 1000.
 *
 * If this device has a board number, then we make a "fake"
 * device called OBP_SYSBOARD that we insert in between the
 * Device and it's logical/current Master.
 */
static void OBPsetBoard(DevInfo, BoardNum, TreePtr, SearchNames)
    DevInfo_t		       *DevInfo;
    int				BoardNum;
    DevInfo_t		      **TreePtr;
    char		      **SearchNames;
{
    static char			BoardName[BUFSIZ];
    register DevInfo_t	       *DevPtr;
    register DevInfo_t	       *SaveDev;
    register DevInfo_t	       *LastDev;
    register DevInfo_t	       *OldMaster = NULL;
    register DevInfo_t	       *Master = NULL;
    DevDefine_t		       *DevDefine = NULL;

    if (BoardNum < 0 || !DevInfo)
	return;

    (void) sprintf(BoardName, "%s%d", OBP_SYSBOARD, BoardNum);

    /*
     * Find the Old Master
     */
    if (DevInfo->Master) {
	OldMaster = FindDeviceByNodeID(DevInfo->Master->NodeID, 
				       *TreePtr);
	if (!OldMaster)
	    OldMaster = FindDeviceByName(DevInfo->Master->Name, 
					 *TreePtr);
    }

    /*
     * If the Master (sysboard) device doesn't exit, create one.
     */
    if (!(Master = FindDeviceByName(BoardName, *TreePtr))) {
	Master = NewDevInfo(NULL);
	Master->Name = strdup(BoardName);
	Master->Unit = BoardNum;
	if (DevDefine = DevDefGet(OBP_SYSBOARD, 0, 0)) {
	    Master->Model = DevDefine->Model;
	    Master->ModelDesc = DevDefine->Desc;
	    Master->Type = DevDefine->Type;
	}
	if (OldMaster) {
	    Master->MasterName = OldMaster->Name;
	    Master->Master = OldMaster;
	}
	AddDevice(Master, TreePtr, SearchNames);
    }

    /*
     * Insert sysboard between Device and it's logical Master.
     */
    if (OldMaster) {
	Master->Master = OldMaster;

	/*
	 * Find Device in OldMaster and remove it.
	 */
	SaveDev = NULL;
	for (DevPtr = OldMaster->Slaves, LastDev = DevPtr; DevPtr; 
	     LastDev = DevPtr, DevPtr = DevPtr->Next) {

	    if (!EQ(DevPtr->Name, DevInfo->Name))
		continue;

	    SaveDev = DevPtr;
	    LastDev->Next = DevPtr->Next;
	    break;
	}

	/*
	 * Add SaveDev back to new Master
	 */
	SaveDev->Next = NULL;
	if (Master->Slaves) {
	    for (DevPtr = Master->Slaves; DevPtr && DevPtr->Next;
		 DevPtr = DevPtr->Next);
	    DevPtr->Next = SaveDev;
	} else
	    Master->Slaves = SaveDev;
    }

    /*
     * Update Device
     */
    if (Master) {
	DevInfo->MasterName = Master->Name;
	DevInfo->Master = Master;
    }
}

/*
 * Decode memory string into number of MBYTES.
 */
static int OBPGetMemSize(MemStr)
    char		       *MemStr;
{
    register char	       *cp;
    char		       *MemBuff;
    u_long			Amt = 0;

    if (strchr(MemStr, '.')) {
	MemBuff = strdup(MemStr);
	for (cp = strtok(MemBuff, "."); cp; cp = strtok((char *)NULL, "."))
	    Amt += strtol(cp, (char **) NULL, 0);
	(void) free(MemBuff);
    } else
	Amt = strtol(MemStr, (char **) NULL, 0);

    return(Amt / MBYTES);
}

/*
 * Decode memory string into memory groups and add as a dev description.
 */
static void OBPSetMemGrps(DevInfo, MemStr)
    DevInfo_t		       *DevInfo;
    char		       *MemStr;
{
    static char			Buff[BUFSIZ];
    char			Str[100];
    register char	       *cp;
    char		       *MemBuff;
    int				GrpNum = 0;

    if (!strchr(MemStr, '.'))
	return;

    Buff[0] = CNULL;
    MemBuff = strdup(MemStr);
    for (cp = strtok(MemBuff, "."); cp; cp = strtok((char *)NULL, ".")) {
	(void) sprintf(Str, "#%d=%s", GrpNum++, 
		       GetSizeStr((u_long) strtol(cp, (char **) NULL, 0), 
				  BYTES));
	if (Buff[0] == CNULL)
	    strcpy(Buff, Str);
	else {
	    strcat(Buff, ", ");
	    strcat(Buff, Str);
	}
    }
    (void) free(MemBuff);

    AddDevDesc(DevInfo, Buff, "Memory Groups", DA_APPEND);
}

/*
 * Probe an unknown entity using OBP.
 */
extern DevInfo_t *OBPprobeUnknown(NodeID, ParentID, OBPtable, 
				  TreePtr, SearchNames)
    OBPnodeid_t			NodeID;
    OBPnodeid_t			ParentID;
    OBPtable_t		       *OBPtable;
    DevInfo_t		      **TreePtr;
    char		      **SearchNames;
{
    register OBPtable_t	       *tptr;
    DevInfo_t		       *DevInfo = NULL;
    register char	      **cpp;
    static char			buff[BUFSIZ];
    int				DevID = -1;
    int				BoardNum = -1;
    char		       *NodeName = NULL;
    char		       *Part = NULL;
    Define_t		       *PartInfo = NULL;
    int				MemSize = 0;

    /*
     * First see if we can find this device by it's OBP node ID
     */
    if (NodeID && TreePtr)
	DevInfo = FindDeviceByNodeID(NodeID, *TreePtr);

    for (tptr = OBPtable; tptr; tptr = tptr->obpt_next) {
	if (!tptr->obpt_key[0])
	    break;

	if (EQ(OBPK_NAME, tptr->obpt_key))
	    NodeName = tptr->obpt_value;
    }

    /*
     * Try to find the device by name.   This usually only
     * works for the root "system" node and a few others.
     */
    if (!DevInfo && NodeName && TreePtr)
	DevInfo = FindDeviceByName(NodeName, *TreePtr);

    /*
     * Didn't find it.
     */
    if (!DevInfo)
	return((DevInfo_t *)NULL);

    /*
     * Get descriptive and other info
     */
    for (tptr = OBPtable; tptr; tptr = tptr->obpt_next) {
	if (!tptr->obpt_key[0])
	    break;

	/*
	 * Skip these
	 */
	if (EQ(tptr->obpt_key, OBPK_NAME))
	    continue;
	else if (EQ(tptr->obpt_key, OBPK_MODEL))
	    Part = tptr->obpt_value;
	else if (EQ(tptr->obpt_key, OBPK_BOARD))
	    BoardNum = strtol(tptr->obpt_value, (char **) NULL, 0);
	else if (EQ(tptr->obpt_key, OBPK_DEVICEID))
	    DevID = strtol(tptr->obpt_value, (char **) NULL, 0);
	else if (EQ(tptr->obpt_key, OBPK_SIZE) && 
		 EQ(NodeName, OBPK_MEMUNIT)) {
	    MemSize = OBPGetMemSize(tptr->obpt_value);
	    OBPSetMemGrps(DevInfo, tptr->obpt_value);
	} else if (EQ(tptr->obpt_key, OBPK_KEYBOARD))
	    DevInfo->Slaves = ProbeKbd((DevInfo_t **)NULL);

	OBPsetOBPinfo(tptr, DevInfo);
    }

    /*
     * Handle special cases
     */
    if (MemSize && DevInfo->Model) {
	(void) sprintf(buff, "%d MB %s", MemSize, DevInfo->Model);
	DevInfo->Model = strdup(buff);
    }
    if (BoardNum >= 0)
	OBPsetBoard(DevInfo, BoardNum, TreePtr, SearchNames);

    /*
     * Add information to existing Device.
     */

    if (Part) {
	PartInfo = OBPGetPart(Part);
	if (PartInfo) {
	    DevInfo->Model = PartInfo->ValStr1;
	    AddDevDesc(DevInfo, PartInfo->ValStr2, NULL, DA_INSERT|DA_PRIME);
	} else if (DevInfo->Model) {
	    (void) sprintf(buff, "%s (%s)", DevInfo->Model, Part);
	    DevInfo->Model = strdup(buff);
	} else
	    DevInfo->Model = strdup(Part);
    }

    return((DevInfo_t *)NULL);
}

/*
 * Check over an OBP node.
 */
static char *OBPIOcheckNode(NodeID, obp_fd, ParentName, ParentID, 
			    TreePtr, SearchNames)
    OBPnodeid_t			NodeID;
    int				obp_fd;
    char		       *ParentName;
    OBPnodeid_t			ParentID;
    DevInfo_t		      **TreePtr;
    char		      **SearchNames;
{
    OBPtable_t		       *OBPtable;
    register OBPtable_t	       *tptr;
    static int			first = TRUE;
    static char			MyName[BUFSIZ];
    static char			buff[BUFSIZ];
    DevDefine_t		       *DevDefine = NULL;
    DevInfo_t		       *DevInfo = NULL;
    DevInfo_t		       *master = NULL;

    /*
     * Get all the properties for the current node.
     */
    if (!(OBPtable = OBPIOgetproptab(obp_fd)))
	return((char *)NULL);

    /*
     * Iterate over all the properties looking for interesting properties
     * or until we find the device type.  If the device type is one which
     * we found in our probe list, go ahead and probe it.
     */
    MyName[0] = CNULL;
    for (tptr = OBPtable; tptr; tptr = tptr->obpt_next) {
	/*
	 * Have we reached the end of the valid list?
	 */
	if (!tptr->obpt_key[0])
	    break;

	if (Debug) printf("OBPIOcheckNode:     NodeID %d <%s> = <%s> (0x%x)\n",
			  NodeID,
			  tptr->obpt_key, tptr->obpt_value,
			  atol(tptr->obpt_value));

	/*
	 * If this is the first node and this property is the Clock Frequency
	 * then save it for later use by the individual CPUs.
	 */
	if (first && !CpuClockFreq && EQ(tptr->obpt_key, OBPK_CLOCKFREQ))
	    CpuClockFreq = strtol(tptr->obpt_value, (char **)NULL, 0);

	if (EQ(tptr->obpt_key, OBPK_DEVTYPE))
	    DevDefine = DevDefGet(tptr->obpt_value, 0, 0);

	if (EQ(tptr->obpt_key, OBPK_NAME) && tptr->obpt_value[0])
	    (void) strcpy(MyName, tptr->obpt_value);
    }

    if (Debug) 
	printf("OBPIOcheckNode: NodeID %d Name = <%s> DevDefine.Name = <%s>\n",
	       NodeID, 
	       (MyName[0]) ? MyName : "", 
	       (DevDefine && DevDefine->Name) ? DevDefine->Name : "");

    /*
     * We use DevDefine->Probe for our own internal (obp.c) use so we
     * need to make sure we only call the right Probe functions
     */
    if (DevDefine && DevDefine->Probe == OBPprobeCPU)
	DevInfo = (*DevDefine->Probe)(NodeID, ParentID, OBPtable, TreePtr,
				      SearchNames);

    if (!DevInfo)
	DevInfo = OBPprobeUnknown(NodeID, ParentID, OBPtable, 
				  TreePtr, SearchNames);

    if (DevInfo) {
	/*
	 * Find or create of Master
	 */
	if (TreePtr && !DevInfo->Master) {
	    master = FindDeviceByNodeID(ParentID, *TreePtr);
	    if (!master && ParentName) {
		master = NewDevInfo(NULL);
		master->Name = strdup(ParentName);
		master->NodeID = ParentID;
	    }
	    DevInfo->Master = master;
	}
	DevInfo->NodeID = NodeID;
	AddDevice(DevInfo, TreePtr, SearchNames);
    }

    first = FALSE;

    return((DevInfo) ? DevInfo->Name : ((MyName[0]) ? MyName : (char *)NULL));
}

/*
 * Traverse OBP I/O nodes
 */
static int OBPIOtraverse(NodeID, obp_fd, ParentName, ParentID, 
			 TreePtr, SearchNames)
    OBPnodeid_t		        NodeID;
    int				obp_fd;
    char		       *ParentName;
    OBPnodeid_t		        ParentID;
    DevInfo_t		      **TreePtr;
    char		      **SearchNames;
{
    OBPnodeid_t		        NewID;
    char		       *MyName;
    char		       *cp;

    if (cp = OBPIOcheckNode(NodeID, obp_fd, ParentName, ParentID, 
			    TreePtr, SearchNames))
	MyName = strdup(cp);

    if (NewID = OBPIOnext(NodeID, obp_fd, OPROMCHILD))
	OBPIOtraverse(NewID, obp_fd, MyName, NodeID, TreePtr, SearchNames);

    if (NewID = OBPIOnext(NodeID, obp_fd, OPROMNEXT))
	OBPIOtraverse(NewID, obp_fd, ParentName, ParentID, 
		      TreePtr, SearchNames);

    return(0);
}

/*
 * Build devices based on what we find using raw I/O calls
 * to the OBP.
 */
static int OBPIObuild(TreePtr, SearchNames)
    DevInfo_t		      **TreePtr;
    char		      **SearchNames;
{
    int				obp_fd;
    int				status;

    /*
     * Note: Only one process at a time can open _PATH_OPENPROM.
     */
    if ((obp_fd = open(_PATH_OPENPROM, O_RDONLY)) < 0) {
	if (Debug) Error("Cannot open \"%s\": %s.", _PATH_OPENPROM, SYSERR);
	return(-1);
    }

    status = OBPIOtraverse(OBPIOnext((OBPnodeid_t) 0, obp_fd, OPROMNEXT),
			   obp_fd, (char *)NULL, 0, TreePtr, SearchNames);
    (void) close(obp_fd);

    return(status);
}

/*
 * Get the "ROM Version" by querying the OBP
 */
extern char *OBPgetVersion()
{
    static char			Version[OPROMMAXPARAM];
#if	defined(OPROMGETVERSION)
    int				obp_fd;
    struct openpromio	       *op;
    static OBPio_t		opio;

    /*
     * Note: Only one process at a time can open _PATH_OPENPROM.
     */
    if ((obp_fd = open(_PATH_OPENPROM, O_RDONLY)) < 0) {
	if (Debug) Error("Cannot open \"%s\": %s.", _PATH_OPENPROM, SYSERR);
	return((char *) NULL);
    }

    op = &(opio.opio_oprom);
    op->oprom_size = sizeof(opio.opio_buff);

    if (ioctl(obp_fd, OPROMGETVERSION, op) < 0) {
	if (Debug) Error("OBP ioctl OPROMGETVERSION failed: %s", SYSERR);
	return((char *) NULL);
    }
    (void) close(obp_fd);

    strcpy(Version, op->oprom_array);
#endif	/* OPROMGETVERSION */

    return((Version[0]) ? Version : (char *) NULL);
}

/*
 * Find and build devices based on what we find from the OBP
 * both from the kernel and the OPENPROM itself.
 */
extern int BuildOpenPROM(TreePtr, SearchNames)
    DevInfo_t		      **TreePtr;
    char		      **SearchNames;
{
    int				found = 1;

    if (OBPKbuild(TreePtr, SearchNames) == 0)
	found = 0;

    if (OBPIObuild(TreePtr, SearchNames) == 0)
	found = 0;

    return(found);
}

/*
 * Get the CPU type by reading and returning the name of the
 * root (top) OBP node from the kernel.
 */
static char *OBPKgetCpuType()
{
    OBPdev_info_t	       *Root;
    static char			Name[BUFSIZ];

    if (!(Root = OBPKgetRoot()))
	return((char *) NULL);
    
    if (!Root->devi_name) {
	if (Debug) Error("OBP Root node does not have a name.");
	return((char *) NULL);
    }

    /*
     * Read the name from kernel space.
     */
    if (KVMget(kd, (u_long) Root->devi_name, (char *) Name, 
	       sizeof(Name), KDT_STRING)) {
	if (Debug) Error("Cannot read OBP root device name.");
	return((char *) NULL);
    }

    if (Name && *Name)
	return(Name);
    else
	return((char *) NULL);
}

/*
 * Get system model type from OBP directly.
 */
static char *OBPIOgetCpuType()
{
    int				obp_fd;
    char		       *Name;

    /*
     * Note: Only one process at a time can open _PATH_OPENPROM.
     */
    if ((obp_fd = open(_PATH_OPENPROM, O_RDONLY)) < 0) {
	if (Debug) Error("Cannot open \"%s\": %s.", _PATH_OPENPROM, SYSERR);
	return((char *) NULL);
    }

    Name = OBPIOcheckNode(OBPIOnext((OBPnodeid_t) 0, obp_fd, OPROMNEXT),
			  obp_fd, (char *) NULL, (OBPnodeid_t) 0, 
			  (DevInfo_t *) NULL, (char **)NULL);

    (void) close(obp_fd);

    return(Name);
}

/*
 * Get system CPU type from OBP.
 */
extern char *OBPGetCpuType()
{
    char		       *Name;

    /*
     * First try getting it from the kernel, if that
     * fails, then query the OBP directly.
     */
    Name = OBPKgetCpuType();
    if (!Name)
	Name = OBPIOgetCpuType();

    if (Name)
	return(OBPcleanName(Name, TRUE));
    else
	return((char *) NULL);
}

#endif	/* HAVE_OPENPROM */
